from typing import List, Union, Tuple, Any, Dict, no_type_check
import svg
from .exon import element_xy, Element, Exon, Variant
import exonviz.exon
import textwrap


# Options for drawing the figure. Used to create the cli parser and default dict
_config = [
    ("width", 9999999, "Maximum width of the figure"),
    ("height", 20, "Exon height"),
    ("noncoding", False, "Show non coding regions"),
    ("gap", 5, "Gap between the exons"),
    ("color", "#4C72B7", "Color for the exons (e.g. 'purple')"),
    ("exonnumber", False, "Show exon number"),
    ("firstexon", 1, "The first exon to draw"),
    ("lastexon", 9999, "The last exon to draw"),
]

config = {key: value for key, value, description in _config}


def shift(
    points: List[Union[float, int]], x_offset: float, y_offset: float
) -> List[Union[float, int]]:
    """Shift the x- and y position by the supplied offsets"""
    return [x + y_offset if i % 2 else x + x_offset for i, x in enumerate(points)]


def bottom_right(elements: List[Element]) -> Tuple[int, int]:
    x = 0
    y = 0
    for e in elements:
        e_x, e_y = element_xy(e)
        x = max(x, e_x)
        y = max(y, e_y)
    return x, y


def draw_legend(
    exons: List[Exon], width: int, height: int, y: float = 0
) -> List[Element]:
    """Draw the legend for variants in Exons"""

    def guess_width(name: str, height: int) -> float:
        """Guess how wide the legend for this variant will be"""
        letter_width = 10
        return height * 1.5 + len(name) * letter_width

    def get_legend_keys(exons: List[Exon]) -> List[Tuple[str, str]]:
        """Extract the legend keys from the Exon Variants"""
        keys = list()
        for exon in exons:
            for variant in exon.variants:
                key = (variant.color, variant.name)
                if key not in keys:
                    keys.append(key)
        return keys

    elements: List[Element] = list()

    # The x-position where we will draw the first entry of the legend
    x_pos: float = height
    for color, name in get_legend_keys(exons):
        entry_width = guess_width(name, height)
        if entry_width > width:
            raise RuntimeError(f"Variant name {name} too wide to render")
        elif x_pos + entry_width > width:
            y = y + height * 1.5
            x_pos = height
        elements.append(svg.Rect(x=x_pos, y=y, width=height, height=height, fill=color))
        elements.append(
            svg.Text(
                x=x_pos + height * 1.5, y=y + 0.5 * height, text=name, class_=["legend"]
            )
        )
        x_pos += 2 * height + entry_width

    # Add styling for the legend
    elements.append(
        svg.Style(
            text=textwrap.dedent(
                f"""
                .legend {{ font: {height*0.8}px sans-serif; fill: black; dominant-baseline: central; }}
                """
            )
        )
    )

    return elements


@no_type_check
def draw_exons(
    exons: List[Exon],
    config: Dict[str, Any],
) -> svg.SVG:
    width = config["width"]
    height = config["height"]

    elements = exonviz.exon.draw_exons(
        exons, width=width, height=height, gap=config["gap"]
    )
    # How far down the page did we go?
    x, y = bottom_right(elements)
    elements += draw_legend(exons, y=y + height, height=height, width=width)

    # Set style for exonnumber, even if we don't need it
    elements.append(
        svg.Style(
            text=textwrap.dedent(
                f"""
            .exonnr {{ text-anchor: middle; dominant-baseline: central; font: {config["height"]/2}px sans-serif; fill: white;}}
            """
            ),
        )
    )
    # The maximum width we have reached for this picture
    canvas_width, canvas_height = bottom_right(elements)

    return svg.SVG(width=canvas_width, height=canvas_height, elements=elements)



from typing import List

from exonviz.draw import shift


def test_no_shift() -> None:
    points: List[float] = [0, 0, 0, 0, 10]
    assert shift(points, 0, 0) == [0, 0, 0, 0, 10]


def test_shift_x_only() -> None:
    """Shift the X by 10"""
    points: List[float] = [0, 0, 0, 10]
    assert shift(points, 10, 0) == [10, 0, 10, 10]


def test_shift_x_y() -> None:
    """Shift both X and Y by 10"""
    points: List[float] = [0, 0, 0, 10]
    assert shift(points, 10, 8) == [10, 8, 10, 18]



from typing import Dict, Any, List, Tuple, cast
import urllib.request
from urllib.error import HTTPError
import json

from mutalyzer_crossmapper import NonCoding, Genomic
from .exon import Exon, Coding, Variant
from GTGT.range import intersect

import logging

logging.basicConfig()
log = logging.getLogger(__name__)

Range = Tuple[int, int]


def fetch_exons(transcript: str) -> Dict[str, Any]:
    """Fetch transcript information from mutalyzer"""

    url = f"https://mutalyzer.nl/api/normalize/{transcript}"

    try:
        response = urllib.request.urlopen(url)
    except HTTPError as e:
        msg = f"Fetching '{url}' returned {e}"
        raise RuntimeError(msg)
    else:
        js = json.loads(response.read())

    if "selector_short" not in js:
        msg = f"No exons found for {transcript} (is it a genomic variant?)"
        raise RuntimeError(msg)
    selector: Dict[str, Any] = js["selector_short"]
    return selector


def fetch_variants(transcript: str) -> Dict[str, Any]:
    """Fetch variant view from mutalyzer"""

    url = f"https://mutalyzer.nl/api/view_variants/{transcript}"

    try:
        response = urllib.request.urlopen(url)
    except HTTPError as e:
        msg = f"Fetching '{url}' returned {e}"
        raise RuntimeError(msg)
    else:
        js: Dict[str, Any] = json.loads(response.read())
    return js


def convert_mutalyzer_range(start: str, end: str, reverse: bool) -> Tuple[int, int]:
    """Convert a mutalyzer range to 0-based coordinates"""
    crossmap = Genomic()

    g_start = crossmap.genomic_to_coordinate(int(start))
    g_end = crossmap.genomic_to_coordinate(int(end))
    if reverse:
        return g_end, g_start + 1
    else:
        return g_start, g_end + 1


def convert_coding_positions(positions: List[List[str]], reverse: bool) -> Range:
    return convert_mutalyzer_range(positions[0][0], positions[0][1], reverse)


def is_reverse(start: str, end: str) -> bool:
    """Is the transcript in reverse"""
    return int(start) > int(end)


def convert_exon_positions(
    positions: List[List[str]], reverse: bool
) -> List[Tuple[int, int]]:
    """Convert exon positions from Mutalyzer to a list of Ranges

    This function also accounts for reverse transcripts
    """
    if reverse:
        return [convert_mutalyzer_range(x[0], x[1], reverse) for x in positions[::-1]]
    else:
        return [convert_mutalyzer_range(x[0], x[1], reverse) for x in positions]


def make_coding(exon: Range, coding_region: Range, start_phase: int) -> Coding:
    """Create the coding region"""
    c = intersect(exon, coding_region)
    if not c:
        return Coding()

    assert len(c) == 1
    # Determine the coding start and end, relative to the exon start
    coding_start, coding_end = c[0]
    coding_start -= exon[0]
    coding_end -= exon[0]

    end_phase = (start_phase + (coding_end - coding_start)) % 3
    return Coding(
        start=coding_start, end=coding_end, start_phase=start_phase, end_phase=end_phase
    )


def exon_variant(exons: List[List[str]], variant: Dict[str, Any]) -> bool:
    """Determine if a given variant falls in an exon"""
    reverse = is_reverse(exons[0][0], exons[0][1])

    x = NonCoding(convert_exon_positions(exons, reverse), reverse)

    position, exon_offset, transcript_offset = x.coordinate_to_noncoding(
        variant["start"]
    )

    return not exon_offset


def parse_view_variants(
    exons: List[List[str]], payload: List[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    """Extract only the exonic variants from the mutalyzer view_variants API payload"""
    # Exclude the regions that flank the variants
    variants = [x for x in payload if x["type"] == "variant"]

    for var in variants:
        # Exclude variants that do not fall inside an exon
        if not exon_variant(exons, var):
            continue
        var["start"], var["end"] = variant_to_ranges(exons, var["start"], var["end"])

    return variants


def inside(exon: Range, variant: Dict[str, Any]) -> bool:
    pos = variant["start"]
    return cast(bool, pos >= exon[0] and pos < exon[1])


def exon_variants(exon: Range, variants: List[Dict[str, Any]]) -> List[Variant]:
    """Create a list of Variants that fall within the exon

    The position of the variants is relative to the Exon start position
    """
    vars = list()
    for var in variants:
        if inside(exon, var):
            relative_position = var["start"] - exon[0]
            vars.append(Variant(relative_position, var["description"], "red"))
    return vars


def exons_to_ranges(exons: List[List[str]], cds: List[str]) -> List[Tuple[int, int]]:
    """Convert mutalyzer exons to python ranges, for both strands"""
    reverse = is_reverse(exons[0][0], exons[0][1])

    x = NonCoding(convert_exon_positions(exons, reverse), reverse)
    g = Genomic()

    output_exons = list()
    for exon in exons:
        e_start = (
            x.coordinate_to_noncoding(g.genomic_to_coordinate(int(exon[0])))[0] - 1
        )
        e_end = x.coordinate_to_noncoding(g.genomic_to_coordinate(int(exon[1])))[0]
        output_exons.append((e_start, e_end))

    return output_exons


def cds_to_ranges(exons: List[List[str]], cds: List[str]) -> Tuple[int, int]:
    """Convert mutalyzer exons to python ranges, for both strands"""
    reverse = is_reverse(cds[0], cds[1])

    x = NonCoding(convert_exon_positions(exons, reverse), reverse)
    g = Genomic()

    cds_start = x.coordinate_to_noncoding(g.genomic_to_coordinate(int(cds[0])))[0] - 1
    cds_end = x.coordinate_to_noncoding(g.genomic_to_coordinate(int(cds[1])))[0]

    return (cds_start, cds_end)


def variant_to_ranges(
    exons: List[List[str]], var_start: int, var_end: int
) -> Tuple[int, int]:
    """Convert mutalyzer exons to python ranges, for both strands"""
    reverse = is_reverse(exons[0][0], exons[0][1])

    x = NonCoding(convert_exon_positions(exons, reverse), reverse)
    g = Genomic()

    new_start = g.genomic_to_coordinate(x.coordinate_to_noncoding(var_start)[0])
    new_end = g.genomic_to_coordinate(x.coordinate_to_noncoding(var_end)[0])
    if reverse:
        return (new_end, new_start)
    else:
        return (new_start, new_end)


def rewrite_reverse_variants(view_variants: Dict[str, Any]) -> None:
    if not view_variants.get("inverted"):
        return

    seq_length = view_variants["seq_length"]
    for view in view_variants["views"]:
        view["start"] = seq_length - view["start"] - 1
        view["end"] = seq_length - view["end"] - 1


def build_exons(
    mutalyzer: Dict[str, Any], view_variants: Dict[str, Any], config: Dict[str, Any]
) -> List[Exon]:
    """Build Exons from the mutalyzer payload"""
    Exons: List[Exon] = list()

    exons = mutalyzer["exon"]["g"]
    cds = mutalyzer["cds"]["g"][0]
    rewrite_reverse_variants(view_variants)
    vars = view_variants["views"]

    # Convert to ranges
    exon_ranges = exons_to_ranges(exons, cds)
    cds_ranges = cds_to_ranges(exons, cds)

    start_phase = 0

    # Used for the exon name
    index = 0

    # Get the variants
    variants = parse_view_variants(exons, vars)

    for exon in exon_ranges:
        # Determine the name of this exon
        index += 1
        name = f"{index}" if config["exonnumber"] else ""

        # Determine the start and end for this exon
        e_start, e_end = exon
        # Determine the coding region for this exon
        coding = make_coding(exon, cds_ranges, start_phase)
        # Determine the variants for this exon
        vars = exon_variants(exon, variants)
        # Determine the size for this exon
        e_size = e_end - e_start
        # Get the color from the configuration
        color = config["color"]

        E = Exon(size=e_size, coding=coding, variants=vars, name=name, color=color)

        # Set the start phase for the next exon
        start_phase = coding.end_phase

        if not config["noncoding"]:
            E.remove_noncoding()
        Exons.append(E)

    return Exons



from dataclasses import dataclass
from typing import Any, List, Optional, Sequence, Dict, Tuple, no_type_check, Union
import copy

import sys

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from _io import TextIOWrapper
from svg import Rect, Polygon, Text, Style

from GTGT.range import intersect

import logging

logging.basicConfig(level="DEBUG")
log = logging.getLogger(__name__)
Element: TypeAlias = Union[Rect, Polygon, Text, Style]


@dataclass()
class Coding:
    start: int = 0
    end: int = 0
    start_phase: int = 0
    end_phase: int = 0

    def __bool__(self) -> bool:
        return self.end > self.start

    @property
    def size(self) -> int:
        return self.end - self.start

    def split(self, size: int) -> "Coding":
        old = (self.start, self.end)
        new = intersect(old, (0, size))
        # Determine the start and end postions for the new Coding region
        if not new:
            new_start = 0
            new_end = 0
        elif len(new) == 1:
            new_start, new_end = new[0]
        else:
            raise RuntimeError
        # Update the start/end of self
        self.start = max(0, self.start - size)
        self.end = max(0, self.end - size)

        # If the old Coding region is empty
        if not self.end:
            # Lift over the old start/end phase to the new
            new_start_phase = self.start_phase
            new_end_phase = self.end_phase
            # Set the old start/end phase to zero
            self.start_phase = 0
            self.end_phase = 0
        # If the old Coding region is not empty, we set the phase of the
        # split region to -1 on either size, so no cap is drawn
        else:
            new_start_phase = self.start_phase
            new_end_phase = 0
            self.start_phase = 0

        return Coding(
            start=new_start,
            end=new_end,
            start_phase=new_start_phase,
            end_phase=new_end_phase,
        )


@dataclass()
class Variant:
    position: int
    name: str
    color: str

    def tsv(self, sep: str = "\t") -> str:
        """Dump an exon as tsv"""
        records = [
            self.position,
            self.name,
            self.color,
        ]
        return sep.join(map(str, records))


class Exon:
    def __init__(
        self,
        size: int,
        coding: Optional[Coding] = None,
        variants: Optional[Sequence[Variant]] = None,
        name: str = "",
        color: str = "#4C72B7",
    ) -> None:
        self.size = size
        self.name = name
        self.color = color

        if coding is None:
            self.coding = Coding()
        else:
            self.coding = coding

        if variants is None:
            self.variants: Sequence[Variant] = list()
        else:
            self.variants = variants

    def __bool__(self) -> bool:
        return self.size != 0

    def __repr__(self) -> str:
        return (
            f"Exon(size={self.size}, "
            f"coding={self.coding}, "
            f"variants={self.variants}, "
            f"name={self.name}, "
            f"color={self.color})"
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Exon):
            raise NotImplementedError
        return all(
            (
                self.size == other.size,
                self.name == other.name,
                self.color == other.color,
                self.coding == other.coding,
                self.variants == other.variants,
            )
        )

    def draw_size(self, scale: float) -> float:
        """Determine how big the Exon is when drawn"""
        return self.size * scale

    def remove_noncoding(self) -> None:
        """Update an exon to remove the non coding region"""
        if not self.coding:
            self.size = 0

        # The new size is the coding size
        self.size = self.coding.size

        # Remove variants outside coding region
        self.variants = [
            v
            for v in self.variants
            if v.position >= self.coding.start and v.position < self.coding.end
        ]
        # Update variant positions
        for v in self.variants:
            v.position -= self.coding.start

        # The coding region starts at 0, and ends at coding size
        self.coding.end = self.coding.end - self.coding.start
        self.coding.start = 0

    def draw(
        self, height: float = 20, scale: float = 1, x: float = 0, y: float = 0
    ) -> List[Element]:
        """Draw the Exon, in SVG format

        Returns a list of SVG elements
        """
        elements: List[Any] = list()

        # If the coding size is not the entire exon
        if self.coding.size != self.size:
            elements.append(self._draw_noncoding(height=height, scale=scale, x=x, y=y))
        if self.coding:
            elements.append(self._draw_coding(height=height, scale=scale, x=x, y=y))
        elements += self._draw_variants(height=height, scale=scale, x=x, y=y)
        if self.name:
            elements.append(self._draw_name(height=height, scale=scale, x=x, y=y))

        return elements

    def min_scale(self, height: float = 20) -> float:
        """Determine the minimum scale this exon can be drawn at"""
        # If there is no coding region, the exon can be drawn at any scale
        if not self.coding:
            return 0.0

        # Total caps we need to draw. If the exon is too small, we raise an error
        total_caps: float = 0
        if self.coding.start_phase != 0:
            total_caps += height * 0.25
        if self.coding.end_phase != 0:
            total_caps += height * 0.25

        scale = total_caps / self.coding.size
        return scale

    def _draw_noncoding(
        self, height: float = 20, scale: float = 1, x: float = 0, y: float = 0
    ) -> Rect:
        """
        Draw the non coding region of the Exon

        If the coding region only overlaps the start or end of the exon, we
        have to draw the non-coding region smaller, so it doesn't stick
        out beyond the arrow/notches
        """
        # Constant values for the drawing
        draw_height = height * 0.5
        y_pos = y + height * 0.25

        # By default, we draw the non coding region the full size of the exon.
        x_pos = x
        width = self.size * scale

        # If only the start of the exon is coding
        start_coding = (
            self.coding and self.coding.start == 0 and self.coding.end < self.size
        )
        if start_coding:
            x_pos = x + (self.coding.end * scale - 1)
            width = (self.size - self.coding.size) * scale + 1

        # If only the end of the exon is coding
        end_coding = (
            self.coding and self.coding.start > 0 and self.coding.end == self.size
        )
        if end_coding:
            x_pos = x
            width = (self.size - self.coding.size) * scale + 1

        return Rect(x=x_pos, y=y_pos, width=width, height=draw_height, fill=self.color)

    def _draw_coding(
        self, height: float = 20, scale: float = 1, x: float = 0, y: float = 0
    ) -> Polygon:
        # Determine x-coordinate for the coding region start
        cx = x + self.coding.start * scale

        # Calculate the size of the arrow/notch
        cap_size = height * 0.25

        # Total size we need to draw.
        size = self.coding.size * scale

        if scale < self.min_scale(height):
            raise ValueError(f"Coding region {self.coding} is too small to draw")

        # fmt: off
        start: List[List[float]] = [
            [ # Square
                cx, y + height,
                cx, y
            ],
            [ # Notch
                cx, y + height,
                cx + cap_size, y + height/2,
                cx, y
            ],
            [ # Arrow
                cx + cap_size, y + height,
                cx, y + height/2,
                cx + cap_size, y
            ]
        ]

        end: List[List[float]] = [
            [ # Square
                cx + size, y,
                cx + size, y + height,
            ],
            [ # Arrow
                cx + size - cap_size, y,
                cx + size, y + height/2,
                cx + size - cap_size, y + height
            ],
            [ # Notch
                cx + size - cap_size, y,
                cx + size, y,
                cx + size - cap_size, y + height/2,
                cx + size, y + height,
                cx + size - cap_size, y + height,
            ]
        ]
        # fmt: on
        start_phase = self.coding.start_phase
        end_phase = self.coding.end_phase
        return Polygon(
            points=list(start[start_phase] + end[end_phase]), fill=self.color
        )

    def _draw_variants(
        self, height: float = 20, scale: float = 1, x: float = 0, y: float = 0
    ) -> Sequence[Rect]:
        elements = list()
        for variant in self.variants:
            elements.append(
                Rect(
                    x=x + variant.position * scale,
                    y=y,
                    width=height / 20,
                    height=height,
                    fill=variant.color,
                )
            )
        return elements

    def _draw_name(
        self, height: float = 20, scale: float = 1, x: float = 0, y: float = 0
    ) -> Text:
        return Text(
            x=x + (self.size * scale / 2),
            y=y + 0.5 * height,
            class_=["exonnr"],
            text=self.name,
        )

    def split(self, size: int) -> "Exon":
        """Split an new exon of size off from self

        Not that this does not take the drawn_size() into account
        """
        # Size of the new exon
        n_size = min(self.size, size)
        self.size = max(0, self.size - size)

        # Update the coding region
        new_coding = self.coding.split(int(size))

        # Update the name
        new_name = self.name

        # Update the color
        new_color = self.color

        # Update the variants
        new_variants = [v for v in self.variants if v.position <= size]
        # Get the variants that remain
        self.variants = [v for v in self.variants if v.position > size]
        # Update the variant positions
        for v in self.variants:
            v.position -= size
        return Exon(
            size=n_size,
            coding=new_coding,
            name=new_name,
            color=new_color,
            variants=new_variants,
        )

    def tsv(self, sep: str = "\t") -> str:
        """Dump an exon as tsv"""
        records = [
            self.size,
            self.name,
            self.color,
            self.coding.start,
            self.coding.end,
            self.coding.start_phase,
            self.coding.end_phase,
        ]
        return sep.join(map(str, records))


def group_exons(
    exons: List[Exon], height: int, gap: int, width: int, scale: float = 1.0
) -> List[List[Exon]]:
    """Group exons on a page, so that they do not go over width"""
    if not exons:
        return [[]]
    page = list()
    row: List[Exon] = list()

    x: float = 0
    for exon in exons:
        while exon:
            # How much space is left on the page
            space_left = int(width - x)

            # Leave some space to account for the coding caps, if needed
            if space_left < 2 * height:
                page.append(row)
                row = list()
                x = 0
            else:
                new_exon = exon.split(int(space_left / scale))
                row.append(new_exon)
                x += gap + new_exon.draw_size(scale)
    page.append(row)
    return page


def draw_exons(exons: List[Exon], width: int, height: int, gap: int) -> List[Element]:
    x: float = height
    y: float = height
    elements = list()
    # The exons will be modified by grouping them, so we make a copy here
    tmp_exons = copy.deepcopy(exons)

    # Calculate the minimum scale at which we can draw every exon
    scale = max(e.min_scale(height) for e in exons)
    # Use the default scale of 1, if possible
    scale = max(scale, 1)

    for row in group_exons(tmp_exons, width=width, height=height, gap=gap):
        for exon in row:
            elements += exon.draw(height=height, scale=scale, x=x, y=y)
            x += exon.draw_size(scale) + gap
        y += 2 * height
        x = height
    return elements


def parse_coding_region(exon_dict: Dict[Any, Any]) -> None:
    """Extract the coding fields into a Coding object, rewrites exon_dict"""
    # Get the coding values out of the exon dictionary
    coding_dict = dict()
    for coding_field in "coding_start coding_end start_phase end_phase".split():
        if coding_field in exon_dict:
            coding_dict[coding_field] = exon_dict.pop(coding_field)
        else:
            coding_dict[coding_field] = 0

    # Rename start and end
    coding_dict["start"] = coding_dict.pop("coding_start")
    coding_dict["end"] = coding_dict.pop("coding_end")

    exon_dict["coding"] = Coding(**{k: int(v) for k, v in coding_dict.items()})


def parse_variants(exon_dict: Dict[Any, Any]) -> None:
    """Extract the variant fields into a list of Variants, rewrites exon_dict"""
    variants: List[Variant] = list()

    if "variant_pos" not in exon_dict:
        exon_dict["variants"] = variants
        return

    positions = exon_dict.pop("variant_pos").split(",")
    names = exon_dict.pop("variant_name").split(",")
    colors = exon_dict.pop("variant_color").split(",")

    if not len(names) == len(positions) and len(names) == len(colors):
        error_msg = "Please specify an equal number of items for each variant_field"
        raise ValueError(error_msg)

    for pos, name, color in zip(positions, names, colors):
        variants.append(Variant(int(pos) if pos else 0, name=name, color=color))
    exon_dict["variants"] = variants


def exon_from_dict(d: Dict[str, str]) -> Exon:
    """Create an Exon from a dictionary"""
    exon_dict: Dict[Any, Any] = d.copy()
    parse_coding_region(exon_dict)
    parse_variants(exon_dict)
    # Convert exon size to int
    exon_dict["size"] = int(exon_dict["size"])
    return Exon(**exon_dict)


def exons_from_tsv(fin: TextIOWrapper) -> List[Exon]:
    header = next(fin).strip("\n").split("\t")
    expected = [
        "size",
        "name",
        "color",
        "coding_start",
        "coding_end",
        "start_phase",
        "end_phase",
    ]
    if not header == expected:
        raise RuntimeError("Unexpected header in TSV file")

    exons: List[Exon] = list()
    for line in fin:
        d = {k: v for k, v in zip(header, line.strip("\n").split("\t")) if v}
        exons.append(exon_from_dict(d))

    return exons


@no_type_check
def element_xy(element: Element) -> Tuple[float, float]:
    """Determine the furthest x,y coordinates for various svg objects"""
    if isinstance(element, Rect):
        x = element.x + element.width
        y = element.y + element.height
    elif isinstance(element, Text):
        x = element.x
        y = element.y
    elif isinstance(element, Polygon):
        if not element.points:
            x = 0
            y = 0
        else:
            x = max(
                (element.points[i] for i in range(len(element.points)) if not i % 2)
            )
            y = max((element.points[i] for i in range(len(element.points)) if i % 2))
    elif isinstance(element, Style):
        x = 0
        y = 0
    else:
        raise ValueError(element)
    return x, y



import pytest

from exonviz.cli import check_input


# A list of valid inputs for the tool
VALID = [
    "NM_004006.3:c.=",
    "NM_000533.5:c.=",
    "ENST00000358273.9:c.=",
    "NG_012337.3(NM_003002.4):c.274G>T",
]


@pytest.mark.parametrize("input", VALID)
def test_check_input_valid(input: str) -> None:
    """Check that we do not accidentally rewrite correct inputs"""
    assert check_input(input) == input


NO_VARIANT = [
    ("NM_004006.3", "NM_004006.3:c.="),
    ("NM_000533.5", "NM_000533.5:c.="),
    ("ENST00000358273.9", "ENST00000358273.9:c.="),
]


@pytest.mark.parametrize("input, rewrite", NO_VARIANT)
def test_check_input_no_variant(input: str, rewrite: str) -> None:
    """Check that we rewrite it to a valid 'no variant' format"""
    assert check_input(input) == rewrite


NO_VERSION = [
    "NM_004006",
    "NM_000533",
    "ENST00000358273",
]


@pytest.mark.parametrize("input", NO_VERSION)
def test_check_input_no_version(input: str) -> None:
    """Raise an error when there is no version"""
    with pytest.raises(RuntimeError):
        check_input(input)



import pytest

from typing import List, Dict, cast
import copy

from exonviz.exon import (
    Coding,
    Exon,
    Variant,
    group_exons,
    exon_from_dict,
    element_xy,
    Element,
    draw_exons,
)
from GTGT.range import Range
from svg import Rect, Text, Polygon, Style


class TestExon:
    @pytest.fixture
    def default_exon(self) -> Exon:
        """
        ==========
        """
        return Exon(100)

    @pytest.fixture
    def center_coding(self) -> Exon:
        """
        ====||||==
        """
        c = Coding(40, 80)
        return Exon(size=100, coding=c)

    @pytest.fixture
    def all(self) -> Exon:
        """Create an exon with all possible features"""
        vars = [
            Variant(10, "A>T", "red"),
            Variant(30, "C>G", "blue"),
            Variant(80, "G>A", "purple"),
        ]
        c = Coding(start=40, end=80, start_phase=1, end_phase=2)
        return Exon(size=100, coding=c, variants=vars, name="Exon-1", color="yellow")

    def test_default_exon(self, default_exon: Exon) -> None:
        """
        GIVEN a default exon of size 100
        WHEN we don't change anythin
        """
        # THEN the size must be 100
        assert default_exon.size == 100

        # THEN the coding region should be false
        assert not default_exon.coding
        # THEN the coding region should start at 0
        assert default_exon.coding.start == 0

        # THEN the variants should be an empty list
        assert default_exon.variants == list()

        # THEN the default color should be set to DCRT blue
        assert default_exon.color == "#4C72B7"

    def test_boolean_exon_true(self) -> None:
        """
        GIVEN an exon of size one
        THEN the boolean value is True
        """
        e = Exon(size=1)
        assert e

    def test_boolean_exon_false(self) -> None:
        """
        GIVEN an exon of size zero
        THEN the boolean value is False
        """
        zero = Exon(size=0)
        assert not zero

    def test_exon_color(self) -> None:
        """
        GIVEN we create an exon with a set color
        THEN that color must be stored in the object
        """
        e = Exon(size=100, color="red")
        assert e.color == "red"

    def test_draw_exon_non_coding(self, all: Exon) -> None:
        """
        GIVEN a yellow exon
        WHEN we draw the exon
        THEN the color should be set
        """
        non_coding = all._draw_noncoding()
        assert non_coding.fill == "yellow"

    def test_draw_fully_non_coding(self) -> None:
        """
        GIVEN an exon without a coding region
        WHEN we draw the non coding region
        THEN it should be the width of the exon
        """
        e = Exon(size=100)

        non_coding = e._draw_noncoding()
        assert non_coding.width == 100

    def test_draw_non_coding_start_coding(self) -> None:
        """
        GIVEN an exon where the start is coding
        WHEN we draw the non coding region
        THEN the drawing should start 1 pixel into the coding region
        THEN the drawing should extend to the end of the exon
        """
        e = Exon(size=100, coding=Coding(start=0, end=20))

        non_coding = e._draw_noncoding(x=13, y=17)
        assert non_coding.x == 19 + 13
        assert non_coding.width == 81

    def test_draw_non_coding_center_coding(self) -> None:
        """
        GIVEN an exon that is coding in the center
        WHEN we draw the non coding region
        THEN it should be the width of the exon (since we will draw the coding
             region on top of it)
        """
        e = Exon(size=100, coding=Coding(start=30, end=70))

        non_coding = e._draw_noncoding()
        assert non_coding.width == 100

    def test_draw_non_coding_end_coding(self) -> None:
        """
        GIVEN an exon where the end is coding
        WHEN we draw the non coding region
        THEN the drawing should start at the start of the exon
        THEN the drawing should extend 1 pixel into the coding region
        """
        e = Exon(size=100, coding=Coding(start=80, end=100))

        non_coding = e._draw_noncoding()
        assert non_coding.x == 0
        assert non_coding.width == 81

    def test_draw_non_coding_fully_coding(self) -> None:
        """
        GIVEN an exon that is fully coding
        WHEN we draw the exon
        THEN we should not draw the non coding region
        """
        e = Exon(size=100, coding=Coding(start=0, end=100))
        elements = e.draw()
        assert len(elements) == 1

    class TestDrawingScale:
        """Test drawing an Exon whith non-default scale"""

        def test_draw_fully_noncoding_scale(self) -> None:
            """
            GIVEN a noncoding exon of size 100
            WHEN we draw the non coding region with scale 1.2
            THEN the drawing width should be 120
            """
            e = Exon(size=100)

            non_coding = e._draw_noncoding(scale=1.2)
            assert non_coding.x == 0
            assert non_coding.width == 120

        def test_draw_non_coding_start_coding(self) -> None:
            """
            GIVEN an exon where the start is coding
            WHEN we draw the non coding region with scale 1.2
            THEN the drawing should start 1 pixel into the coding region
            THEN the drawing should extend to the end of the exon
            """
            e = Exon(size=100, coding=Coding(start=0, end=20))

            non_coding = e._draw_noncoding(scale=1.2)
            assert non_coding.x == 23
            assert non_coding.width == 97

        def test_draw_non_coding_center_coding(self) -> None:
            """
            GIVEN an exon that is coding in the center
            WHEN we draw the non coding region with scale 1.2
            THEN it should be the width of the exon (since we will draw the coding
                 region on top of it)
            """
            e = Exon(size=100, coding=Coding(start=30, end=70))

            non_coding = e._draw_noncoding(scale=1.2)
            assert non_coding.width == 120

        def test_draw_non_coding_end_coding(self) -> None:
            """
            GIVEN an exon where the end is coding
            WHEN we draw the non coding region with scale 1.2
            THEN the drawing should start at the start of the exon
            THEN the drawing should extend 1 pixel into the coding region
            """
            e = Exon(size=100, coding=Coding(start=80, end=100))

            non_coding = e._draw_noncoding(scale=1.2)
            assert non_coding.x == 0
            assert non_coding.width == 97

        # fmt: off
        coding_region_frames = [
            # start phase, points
            (0, [
                0, 20,
                0, 0,
                120, 0,
                120, 20,
            ]),
            (1, [
                0, 20,
                0, 0,
                115, 0,
                120, 10,
                115, 20
            ]),
            (2, [
                0, 20,
                0, 0,
                115.0, 0,
                120.0, 0,
                115.0, 10.0,
                120.0, 20,
                115, 20
            ])
        ]
        # fmt: on
        @pytest.mark.parametrize("end_phase, points", coding_region_frames)
        def test_draw_coding(self, end_phase: int, points: List[int]) -> None:
            """
            Test drawing the coding region with scale 1.2
            Since the start_phase will be drawn the same, we only vary the end phase
            """
            c = Coding(0, 100, start_phase=0, end_phase=end_phase)
            E = Exon(size=100, coding=c, color="blue")
            coding = E._draw_coding(scale=1.2)

            assert coding == Polygon(points=list(points), fill="blue")

        def test_draw_variants_scale(self, all: Exon) -> None:
            """
            GIVEN an exon with variant
            WHEN the variants are drawn with a scale
            THEN the variant positions should be shifted to match
            """
            variants = all._draw_variants(scale=1.2)

            assert variants[0].x == 12
            assert variants[1].x == 36
            assert variants[2].x == 96

        def test_draw_name_scale(self, all: Exon) -> None:
            """
            GIVEN an exon with a name
            WHEN the name is drawn with a scale
            THEN the name position should be shifted to match the new center of the exon
            """
            text = all._draw_name(height=20, scale=1.2, x=11, y=23)
            # THEN the name must be centered in the middle of the exon
            assert text.x == 11 + 60  # offset + half the exon size
            assert text.y == 23 + 10  # offset + half the height

        def test_draw_coding_tiny_phase_zero(self) -> None:
            """
            GIVEN a tiny coding exon with start and end phase 0
            WHEN we draw it
            THEN there should be no error for the notch/arrow
            """
            E = Exon(size=1, coding=Coding(start=0, end=1, start_phase=0, end_phase=0))
            E._draw_coding(height=20, scale=1)

        def test_draw_coding_tiny_start_phase_one(self) -> None:
            """
            GIVEN a tiny coding exon with start phase 1, end phase 0
            WHEN we draw it
            THEN there should be no error
            """
            size = 5
            E = Exon(
                size=size, coding=Coding(start=0, end=size, start_phase=1, end_phase=0)
            )
            E._draw_coding(height=20, scale=1)

        def test_draw_coding_tiny_end_phase_one(self) -> None:
            """
            GIVEN a tiny coding exon with start phase 0, end phase 1
            WHEN we draw it
            THEN there should be no error
            """
            size = 5
            E = Exon(
                size=size, coding=Coding(start=0, end=size, start_phase=0, end_phase=1)
            )
            E._draw_coding(height=20, scale=1)

        def test_draw_coding_tiny_start_phase_one_end_phase_one(self) -> None:
            """
            GIVEN a tiny coding exon with start phase 1, end phase 1
            WHEN we draw it
            THEN there should be no error
            """
            size = 10
            E = Exon(
                size=size, coding=Coding(start=0, end=size, start_phase=1, end_phase=1)
            )
            E._draw_coding(height=20, scale=1)

    def test_draw_exon_coding_color(self, center_coding: Exon) -> None:
        """
        GIVEN an exon where the center region is coding
        WHEN we draw the coding region
        THEN the color should match
        """
        coding = center_coding._draw_coding()
        assert coding.fill == "#4C72B7"

    # fmt: off
    coding_region_frames = [
        # Format is start_phase, end_phase, list of polygon points
        # (x and y are interleaved)
        # ||||
        (0, 0, [
            0, 20,
            0, 0,
            100, 0,
            100, 20,
        ]),
        # |||>
        (0, 1, [
            0, 20,
            0, 0,
            95, 0,
            100, 10,
            95, 20
        ]),
        # |||<
        (0, 2, [
            0, 20,
            0, 0,
            95, 0,
            100, 0,
            95, 10,
            100, 20,
            95, 20
        ]),
        # >|||
        (1, 0, [
            0, 20,
            5, 10,
            0, 0,
            100, 0,
            100, 20,
        ]),
        # >||>
        (1, 1, [
            0, 20,
            5, 10,
            0, 0,
            95, 0,
            100, 10,
            95, 20
        ]),
        # >||<
        (1, 2, [
            0, 20,
            5, 10,
            0, 0,
            95, 0,
            100, 0,
            95, 10,
            100, 20,
            95, 20
        ]),
        # <|||
        (2, 0, [
            5, 20,
            0, 10,
            5, 0,
            100, 0,
            100, 20,
        ]),
        # <||>
        (2, 1, [
            5, 20,
            0, 10,
            5, 0,
            95, 0,
            100, 10,
            95, 20
        ]),
        # <||<
        (2, 2, [
            5, 20,
            0, 10,
            5, 0,
            95, 0,
            100, 0,
            95, 10,
            100, 20,
            95, 20
        ])
    ]
    # fmt: on
    @pytest.mark.parametrize("start_phase, end_phase, points", coding_region_frames)
    def test_draw_coding(
        self, start_phase: int, end_phase: int, points: List[int]
    ) -> None:
        E = Exon(size=100, coding=Coding(0, 100, start_phase, end_phase), color="blue")
        coding = E._draw_coding()

        assert coding == Polygon(points=list(points), fill="blue")

    @pytest.mark.parametrize("start_phase, end_phase, points", coding_region_frames)
    def test_draw_coding_offset(
        self, start_phase: int, end_phase: int, points: List[int]
    ) -> None:
        E = Exon(size=100, coding=Coding(0, 100, start_phase, end_phase), color="blue")
        # Test setting x or y offset
        x_offset = 8
        y_offset = 13

        coding = E._draw_coding(x=x_offset, y=y_offset)

        # Shift the expected points (x and y are interleaved)
        shifted = list()
        for i, number in enumerate(points):
            if i % 2 == 0:
                shifted.append(number + x_offset)
            else:
                shifted.append(number + y_offset)

        assert coding == Polygon(points=list(shifted), fill="blue")

    def test_draw_coding_too_small(self) -> None:
        """
        GIVEN an exon with a (too) small coding region
        WHEN we try to draw it
        THEN we should get a ValueError
        """
        too_small = Exon(
            size=100, coding=Coding(start=0, end=9, start_phase=1, end_phase=1)
        )
        big_enough = Exon(
            size=100, coding=Coding(start=0, end=10, start_phase=1, end_phase=1)
        )

        with pytest.raises(ValueError):
            too_small._draw_coding()

        assert big_enough._draw_coding()

    def test_draw_exon_offset_noncoding(self, default_exon: Exon) -> None:
        """
        GIVEN a default exon of size 100
        WHEN we draw this exon with an x and y offset
        """
        elements = default_exon.draw(x=11, y=13)
        # THEN there should only be a single element, since the exon is non coding
        assert len(elements) == 1

        non_coding = cast(Rect, elements[0])
        # THEN the drawing should start at the specified x offset
        assert non_coding.x == 11  #  offset

        # THEN the drawing should start at the specified y offset, plus a factor to
        # account for height
        assert non_coding.y == 13 + 5  #  offset + 0.25 * height

    def test_split_exon(self, all: Exon) -> None:
        """
        GIVEN an exon with all features enabled
        WHEN we split this exon in half
        """
        # Split the exon in half
        new = all.split(size=50)

        # THEN the new and old size should both be 50
        assert new.size == 50
        assert all.size == 50

        # THEN the coding region of the new Exon should start at 40
        assert new.coding.start == 40
        # THEN the coding region of the old exon should start at 0
        assert all.coding.start == 0

        # THEN the coding region of the new exon should end at 50
        assert new.coding.end == 50
        # THEN the coding region of the old exon should end at 30
        assert all.coding.end == 30

        # THEN the start phase of the new exon should be 1
        assert new.coding.start_phase == 1
        # THEN the end phase of new exon should be 0, since we don't want to
        # draw a notch or arrow
        assert new.coding.end_phase == 0

        # THEN the start of the old exon should be 0, since we don't want to
        # draw a notch or arrow
        assert all.coding.start_phase == 0
        # THEN the end of the old exon should be 2
        assert all.coding.end_phase == 2

        # THEN both exons should have the original name
        assert new.name == "Exon-1"
        assert all.name == "Exon-1"

        # THEN two of the variants should be in the new exon
        assert new.variants[0].position == 10
        assert new.variants[1].position == 30

        # THEN the position of the variant in the old exon should be updated
        assert all.variants[0].position == 30

    def test_take_full_exon(self, all: Exon) -> None:
        """
        GIVEN an exon with all features enabled
        WHEN we 'split' the exon but take the whole region
        """

        new = all.split(size=100)

        # THEN the new size should be the total size
        assert new.size == 100
        # THEN the coding start and end phase should be unchanged
        assert new.coding.start_phase == 1
        assert new.coding.end_phase == 2

        # THEN the old exon should be empty (false)
        assert not all

    def test_take_full_non_coding(self, default_exon: Exon) -> None:
        """
        GIVEN a default, non coding exon with size 100
        WHEN we 'split' the exon but take the whole region
        """
        new = default_exon.split(100)

        # THEN the size of the new exon should be unchanged
        assert new.size == 100

    def test_take_bigger_exon(self, all: Exon) -> None:
        """
        GIVEN an exon with all features enabled
        WHEN we 'split' the exon but take a region bigger than the exon itself
        """
        new = all.split(size=1000)

        # THEN the new size should be the original size
        assert new.size == 100
        # THEN the left over size should be zero
        assert all.size == 0

        # THEN the left over exon should be false
        assert not all

    def test_split_exon_color(self, all: Exon) -> None:
        """
        GIVEN an exon with all features enabled
        WHEN we 'split' the exon in half
        THEN both exons should keep the color
        """
        new = all.split(size=1000)

        assert new.color == "yellow"
        assert all.color == "yellow"

    def test_draw_name(self) -> None:
        """
        GIVEN an exon with a name
        WHEN the exon is drawn
        """
        e = Exon(size=100, name="Exon-1")

        elements = e.draw(height=10, x=11, y=23)

        assert len(elements) == 2

        # The name is the second element in the drawing
        text = cast(Text, elements[1])

        # THEN the name must be centered in the middle of the exon
        assert text.x == 11 + 50  # offset + half the exon size
        assert text.y == 23 + 5  # offset + half the height
        # THEN the name must be set
        assert text.text == "Exon-1"

    exon_dict = [
        ({"size": "100"}, Exon(100)),
        ({"size": "100", "name": "Exon-1"}, Exon(100, name="Exon-1")),
        ({"size": "100", "color": "blue"}, Exon(100, color="blue")),
        ({"size": "100", "coding_start": 5}, Exon(100, coding=Coding(start=5))),
        ({"size": "100", "coding_end": 5}, Exon(100, coding=Coding(end=5))),
        ({"size": "100", "start_phase": 1}, Exon(100, coding=Coding(start_phase=1))),
        ({"size": "100", "end_phase": 1}, Exon(100, coding=Coding(end_phase=1))),
        ({"size": "100", "end_phase": 1}, Exon(100, coding=Coding(end_phase=1))),
        (
            {
                "size": "100",
                "variant_pos": "10",
                "variant_name": "A>T",
                "variant_color": "red",
            },
            Exon(size=100, variants=[Variant(position=10, name="A>T", color="red")]),
        ),
        (
            {
                "size": "100",
                "variant_pos": "10,22",
                "variant_name": "A>T,stop codon",
                "variant_color": "red,purple",
            },
            Exon(
                size=100,
                variants=[
                    Variant(position=10, name="A>T", color="red"),
                    Variant(position=22, name="stop codon", color="purple"),
                ],
            ),
        ),
    ]

    @pytest.mark.parametrize("d, exon", exon_dict)
    def test_exon_from_dict(self, d: Dict[str, str], exon: Exon) -> None:
        """
        GIVEN an exon dictionary
        WHEN the the dictionary is used to initialise the exon
        THEN the exon must match the expected exon
        """
        assert exon_from_dict(d) == exon

    def test_remove_noncoding_noncoding_exon(self, default_exon: Exon) -> None:
        """
        GIVEN an exon that is non coding
        WHEN the non coding part is removed
        THEN the entire exon should be empty
        """
        default_exon.remove_noncoding()
        assert default_exon.size == 0

    def test_remove_noncoding_region(self, center_coding: Exon) -> None:
        """
        GIVEN an exon that is coding in the center
        WHEN the non coding part is removed
        THEN the entire exon should match the coding region
        """
        center_coding.remove_noncoding()
        assert center_coding.size == 40
        assert center_coding.coding.start == 0
        assert center_coding.coding.end == 40

    def test_remove_noncoding_variants(self) -> None:
        """
        GIVEN an exon with variants inside and outside the coding region
        WHEN the coding region is removed
        THEN the non-coding variants should be removed
        THEN the variant positions should be updated
        """
        c = Coding(40, 80)
        vars = [
            # Before coding
            Variant(10, "A>T", "red"),
            # In coding
            Variant(50, "C>T", "blue"),
            # After coding
            Variant(90, "A>T", "green"),
        ]
        e = Exon(size=100, coding=c, variants=vars)
        e.remove_noncoding()
        # Test that the non coding variants were removed
        # Test that the variant position has been updated from 50 to 10
        assert e.variants == [Variant(10, "C>T", "blue")]

    def test_group_exons_with_variants(self, all: Exon) -> None:
        """
        GIVEN a list of Exons
        WHEN we draw the exons
        THEN the list of exons should be unchanged
        """
        exons = [Exon(100), Exon(200)]
        before = copy.deepcopy(exons)
        # Draw the exons, and discard the drawing
        draw_exons(exons, height=20, gap=10, width=1_000_000)
        assert exons == before


class TestCoding:
    def test_default_coding(self) -> None:
        c = Coding()
        assert c.start == 0
        assert c.end == 0
        assert c.size == 0

        # An empty coding region is false
        assert not c

    split_coding = [
        # size, new, old
        (50, (40, 50), (0, 30)),
        # The size ends before the coding region
        (10, (0, 0), (30, 70)),
        # The size has the entire coding region
        (80, (40, 80), (0, 0)),
        # The size is bigger than the coding region
        (100, (40, 80), (0, 0)),
    ]

    @pytest.mark.parametrize("size, new, old", split_coding)
    def test_split_coding(self, size: int, new: Range, old: Range) -> None:
        """
        GIVEN the coding region c
        WHEN it is split to size "size"
        THEN the new and old ranges should be as specified
        """
        c = Coding(40, 80, 1, 2)
        n = c.split(size=size)

        old_range = (c.start, c.end)
        new_range = (n.start, n.end)

        assert new_range == new
        assert old_range == old

    def test_split_coding_update_phase(self) -> None:
        """
        GIVEN a Coding region has a start and end phase
        WHEN the Coding region is split
        THEN the new Coding should get the start_phase
             the end_phase of the new Coding should be set to -1,
             so no cap is drawn

             the old Coding should keep the end_phase
             the start_phase of the old Coding region should be set to -1,
             so no cap is drawn
        """
        old = Coding(start=40, end=80, start_phase=1, end_phase=2)

        new = old.split(50)

        assert new.start_phase == 1
        assert new.end_phase == 0

        assert old.start_phase == 0
        assert old.end_phase == 2

    def test_split_coding_phase_all(self) -> None:
        """
        GIVEN a Coding region has a start and end phase
        WHEN the Coding region is 'split' into one big chunk
        THEN the Coding start/end phase of new should be lifted over
        """
        old = Coding(start=40, end=80, start_phase=1, end_phase=2)
        new = old.split(1000)

        assert new.start_phase == 1
        assert new.end_phase == 2

        # If the old Coding is empty, the start/end phase should be zero
        assert old.start_phase == 0
        assert old.end_phase == 0


class TestVariant:
    @pytest.fixture
    def with_variant(self) -> Exon:
        vars = [Variant(10, "A>T", "red"), Variant(30, "C>G", "blue")]
        return Exon(size=100, variants=vars)

    def test_draw_variants(self, with_variant: Exon) -> None:
        """
        GIVEN an exon with two variants
        WHEN the exon is drawn
        """
        elements = with_variant.draw(height=20)
        assert len(elements) == 3

        # The first variant is the second element
        var1 = cast(Rect, elements[1])
        # THEN the x position of the variant must be the variant postion
        assert var1.x == 10
        # THEN the color of the variant must be set
        assert var1.fill == "red"

        # The second variant is the third element
        var2 = cast(Rect, elements[2])
        # THEN the x position of the variant must be the variant position
        assert var2.x == 30
        # THEN the color of the variant must be set
        assert var2.fill == "blue"

    def test_draw_variants_offset(self, with_variant: Exon) -> None:
        """
        GIVEN an exon with two variants
        WHEN the exon is drawn with an x offset
        """
        elements = with_variant.draw(height=20, x=11)
        var1 = cast(Rect, elements[1])

        # THEN the x position of the first variant must include
        # the offset
        assert var1.x == 10 + 11  #  variant.position + offset


class TestDrawing:
    to_page = [
        # Exon list, page size, gap, result
        # Empty list
        ([], 100, 0, [[]]),
        # Single exon that fits on one page
        ([Exon(50)], 100, 0, [[Exon(50)]]),
        # Single exon that fits exactly
        ([Exon(100)], 100, 0, [[Exon(100)]]),
        # Single exon that doesn't fit
        ([Exon(100)], 99, 0, [[Exon(99)], [Exon(1)]]),
        # Two exons fit on one row
        ([Exon(50), Exon(50)], 100, 0, [[Exon(50), Exon(50)]]),
        ## Two exons don't fit on one row
        ([Exon(50), Exon(51)], 100, 0, [[Exon(50), Exon(50)], [Exon(1)]]),
        ## Two exons that almost fit on one row
        ([Exon(50), Exon(50)], 99, 0, [[Exon(50), Exon(49)], [Exon(1)]]),
        # Two exons fit on one row, but not with a gap
        ([Exon(50), Exon(50)], 100, 1, [[Exon(50), Exon(49)], [Exon(1)]]),
        # One exon that fits with a giant gap, since we dont include a gap for one exon
        ([Exon(100)], 100, 9999, [[Exon(100)]]),
        # Giant gap, exon doesn't fit
        ([Exon(100)], 99, 9999, [[Exon(99)], [Exon(1)]]),
        # Fits exactly, with gap of 10
        ([Exon(45), Exon(45)], 100, 10, [[Exon(45), Exon(45)]]),
        # Off by one, so it doesn't fit, with gap of 10
        ([Exon(46), Exon(45)], 100, 10, [[Exon(46), Exon(44)], [Exon(1)]]),
        # fmt: off
        # One exon that fits, including a gap of 10
        (
            [Exon(100) for _ in range(1)],
            100,
            10,
            [[Exon(100) for _ in range(1)]]
        ),
        # Two exons that fit, including gaps of size 10
        (
            [Exon(100) for _ in range(2)],
            210,
            10,
            [[Exon(100) for _ in range(2)]]
        ),
        # Ten exons that fit, including gaps of size 10
        (
            [Exon(100) for _ in range(10)],
            1090,
            10,
            [[Exon(100) for _ in range(10)]]
        ),
        # fmt: on
    ]

    @pytest.mark.parametrize("exons, width, gap, page", to_page)
    def test_exons_on_page(
        self, exons: List[Exon], width: int, gap: int, page: List[List[Exon]]
    ) -> None:
        new_page = group_exons(exons, height=20, scale=1, gap=gap, width=width)

        assert new_page == page

    elements = [
        # element, max_x, max_y
        (Rect(x=1, y=2, width=11.0, height=23), 12, 25),
        (Text(x=1, y=1, text="text"), 1, 1),
        (Polygon(points=[0, 0, 10, 5, 0, 15, 20, 25, 13, 17]), 20, 25),
        # Empty polygon
        (Polygon(fill="yellow"), 0, 0),
        # Style for exon number
        (Style(), 0, 0),
    ]

    @pytest.mark.parametrize("element, max_x, max_y", elements)
    def test_element_size(self, element: Element, max_x: float, max_y: float) -> None:
        """Determine the size of various SVG elements"""
        x, y = element_xy(element)

        assert x == max_x
        assert y == max_y

    exon_scale = [
        # Non coding exons can be drawn at any scale
        (Exon(100), 0),
        # A coding region with phase 0 can be drawn at any scale
        (Exon(100, Coding(start=0, end=1, start_phase=0, end_phase=0)), 0),
        # Drawing a single cap takes 5px, so the scale for a coding region of size 1 is 5
        (Exon(100, Coding(start=0, end=1, start_phase=1, end_phase=0)), 5),
        # Drawing two caps, we need a scale of 10
        (Exon(100, Coding(start=0, end=1, start_phase=1, end_phase=1)), 10),
    ]

    @pytest.mark.parametrize("exon, expected_scale", exon_scale)
    def test_determine_exon_scale(self, exon: Exon, expected_scale: float) -> None:
        """
        GIVEN an exon
        WHEN we determine the minimal scale it can be drawn self.assertTrue(
        THEN it has to match the expected scale
        """
        height = 20
        assert exon.min_scale(height=height) == expected_scale
        # Test that there is no value error
        exon.draw(height=height, scale=expected_scale)



"""
Module that contains the command line app, so we can still import __main__
without executing side effects
"""

import argparse
import re
import sys
import gzip
from importlib import resources
from collections import defaultdict

from typing import List, Dict, Any
from .draw import draw_exons
from .exon import Exon, Variant, exons_from_tsv
from .mutalyzer import fetch_exons, fetch_variants, build_exons

from .draw import _config


def get_MANE() -> Dict[str, str]:
    fname = str(resources.files("exonviz.data").joinpath("mane.txt.gz"))
    with gzip.open(fname, "rt") as fin:
        mane = dict()
        for line in fin:
            gene, transcript = line.strip("\n").split("\t")
            mane[gene] = transcript
    return mane


def check_input(transcript: str) -> str:
    """Check the input from the user, and rewrite when needed"""
    # If it looks like there is no variant
    if re.match(r"^\w+\.\d+$", transcript):
        return f"{transcript}:c.="

    # If it looks like the user forgot the version number, there isn't much we can do
    if re.match(r"^\w+$", transcript):
        msg = "Please specify the version of the transcript you are interested in: "
        raise RuntimeError(msg + transcript)

    return transcript


def make_exons(transcript: str, config: Dict[str, Any]) -> List[Exon]:
    """Make or fetch the requested exons"""

    # If the transcript is actually the gene name, substitute the MANE transcript
    MANE = get_MANE()
    transcript = MANE.get(transcript, transcript)
    if transcript in MANE:
        transcript = MANE[transcript]
    else:
        # Does the transcript format make sense?
        transcript = check_input(transcript)

    exons = fetch_exons(transcript)
    variants = fetch_variants(transcript)

    return build_exons(exons, variants, config)


def make_option_parser() -> argparse.ArgumentParser:
    """Create a parser for the drawing options"""
    parser = argparse.ArgumentParser(add_help=False)

    for key, value, description in _config:
        # Booleans are a special case
        if type(value) == bool:
            action = "store_false" if value else "store_true"
            parser.add_argument(
                f"--{key}", default=value, action=action, help=description
            )
        else:
            parser.add_argument(
                f"--{key}",
                type=type(value),
                default=value,
                help=description,
            )

    return parser


def make_parser() -> argparse.ArgumentParser:
    option_parser = make_option_parser()

    parser = argparse.ArgumentParser(
        description="Visualise exons and mutations",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        parents=[option_parser],
    )
    parser.add_argument(
        "--dump-exons", type=str, help="Write exons to the specified file"
    )
    parser.add_argument(
        "--dump-variants", type=str, help="Write variants to the specified file"
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--transcript", help="Transcript (with version) to visualise")
    group.add_argument("--exon-tsv", help="TSV file containing exons")
    parser.add_argument("--variant-tsv", help="TSV file containing variants")

    return parser


def dump_exons(exons: List[Exon], fname: str) -> None:
    """Write the exons to the specified file"""
    header = [
        "size",
        "name",
        "color",
        "coding_start",
        "coding_end",
        "start_phase",
        "end_phase",
    ]
    with open(fname, "wt") as fout:
        print(*header, sep="\t", file=fout)
        for exon in exons:
            print(exon.tsv(sep="\t"), file=fout)


def dump_variants(exons: List[Exon], fname: str) -> None:
    """Write the variants to the specified file"""
    header = [
        "exon",
        "position",
        "name",
        "color",
    ]
    with open(fname, "wt") as fout:
        print(*header, sep="\t", file=fout)
        for i, exon in enumerate(exons, 1):
            for variant in exon.variants:
                print(i, end="\t", file=fout)
                print(variant.tsv(sep="\t"), file=fout)


def exons_from_mutalyzer(transcript: str, config: Dict[str, Any]) -> List[Exon]:
    """Attempt to create exons from mutalyzer"""
    try:
        exons = make_exons(transcript, config)
    except RuntimeError as e:
        print(e, file=sys.stderr)
        exit(1)
    return exons


def exons_from_tsv_file(fname: str) -> List[Exon]:
    """Read Exons from a tsv file"""
    with open(fname) as fin:
        return exons_from_tsv(fin)


def variants_from_tsv_file(fname: str) -> Dict[int, List[Variant]]:
    """Read Variants from a tsv file"""
    variants = defaultdict(list)
    with open(fname) as fin:
        header = next(fin).strip("\n").split("\t")
        assert header == ["exon", "position", "name", "color"]
        for line in fin:
            exon, pos, name, color = line.strip("\n").split("\t")
            variants[int(exon)].append(Variant(int(pos), name, color))
    return variants


def main() -> None:
    parser = make_parser()
    args = parser.parse_args()
    # Make the configuration for the drawing
    config = dict()
    for key, *_ in _config:
        config[key] = getattr(args, key)

    # Create the exons
    if args.transcript:
        exons = exons_from_mutalyzer(args.transcript, config)
    elif args.exon_tsv:
        exons = exons_from_tsv_file(args.exon_tsv)

    if args.variant_tsv:
        vars = variants_from_tsv_file(args.variant_tsv)
        for i, exon in enumerate(exons, 1):
            exon.variants = vars[i]

    if args.dump_exons:
        dump_exons(exons, args.dump_exons)
    if args.dump_variants:
        dump_variants(exons, args.dump_variants)

    else:
        plot = draw_exons(
            exons,
            config=config,
        )
        print(plot)

    if __name__ == "__main__":
        main()



import pytest
from exonviz.mutalyzer import (
    convert_exon_positions,
    convert_coding_positions,
    convert_mutalyzer_range,
    is_reverse,
    make_coding,
    parse_view_variants,
    exon_variants,
    exons_to_ranges,
    variant_to_ranges,
    rewrite_reverse_variants,
    cds_to_ranges,
    exon_variant,
    inside,
    Range,
)
from exonviz.exon import Coding, Variant

from typing import Any, Dict, List, Tuple

# Example mutalyzer payload
mutalyzer = {
    "exon": {
        "g": [["1", "268"], ["269", "330"], ["11284", "13992"]],
    },
    "cds": {"g": [["238", "11295"]]},
}


# def test_convert_coding_positions() -> None:
#     assert convert_coding_positions([["238", "11295"]]) == (237, 11294)
#
#
# def test_convert_coding_positions_reverse() -> None:
#     assert convert_coding_positions([["29199", "7218"]]) == (29198, 7217)

mut_positions = [
    (["1", "268"], False, (0, 268)),
    (["269", "330"], False, (268, 330)),
    (["11284", "13992"], False, (11283, 13992)),
    (["7748", "1"], True, (0, 7748)),
]


@pytest.mark.parametrize("positions, reverse, expected", mut_positions)
def test_convert_mutalyzer_positions(
    positions: List[str], reverse: bool, expected: Tuple[int, int]
) -> None:
    assert convert_mutalyzer_range(positions[0], positions[1], reverse) == expected


def test_is_reverse() -> None:
    assert is_reverse("10", "4")
    assert not is_reverse("4", "10")


def test_convert_mutalyzer_positions_reverse() -> None:
    # Taken from ENST00000436367.6
    positions = [
        ["462349", "462187"],
        ["454236", "454122"],
        ["358790", "358665"],
        ["286796", "286669"],
        ["223577", "223516"],
        ["186998", "186861"],
        ["29359", "27283"],
        ["7748", "1"],
    ]

    expected = [
        (0, 7748),
        (27282, 29359),
        (186860, 186998),
        (223515, 223577),
        (286668, 286796),
        (358664, 358790),
        (454121, 454236),
        (462186, 462349),
    ]

    assert convert_exon_positions(positions, reverse=True) == expected


coding = [
    # Exon, coding_region, start_phase, Coding
    ((0, 10), (20, 30), 0, Coding()),
    ((0, 10), (0, 10), 0, Coding(0, 10, end_phase=1)),
    ((0, 10), (5, 12), 0, Coding(5, 10, end_phase=2)),
    ((0, 10), (-5, 12), 2, Coding(0, 10, start_phase=2, end_phase=0)),
    ((100, 110), (100, 200), 0, Coding(0, 10, end_phase=1)),
]


@pytest.mark.parametrize("exon, coding_region, start_phase, expected", coding)
def test_make_coding(
    exon: Range, coding_region: Range, start_phase: int, expected: Coding
) -> None:
    c = make_coding(exon, coding_region, start_phase)
    assert c == expected


view_variants: Any = [
    ([{"type": "outside"}], list(), list()),
    (
        [
            {"type": "outside"},
            {"type": "variant", "description": "130del", "start": 3615, "end": 3614},
        ],
        [["4952", "4794"], ["3616", "2954"]],
        [
            {"type": "variant", "description": "130del", "start": 160, "end": 159},
        ],
    ),
    (
        [
            {"type": "outside"},
            {"type": "variant", "description": "274G>T", "start": 7124, "end": 7125},
        ],
        [["5027", "5113"], ["6011", "6127"], ["7021", "7165"], ["12959", "13948"]],
        [
            {"type": "variant", "description": "274G>T", "start": 308, "end": 309},
        ],
    ),
    (
        [
            {"type": "outside"},
            {"type": "variant", "description": "53del", "start": 6010, "end": 6011},
        ],
        [["5027", "5113"], ["6011", "6127"], ["7021", "7165"], ["12959", "13948"]],
        [
            {"type": "variant", "description": "53del", "start": 87, "end": 88},
        ],
    ),
    # Non-coding variant in NG_012337.3(NM_003002.4), inside an exon
    (
        [
            {"type": "outside"},
            {"type": "variant", "description": "-10del", "start": 5051, "end": 5052},
        ],
        [["5027", "5113"], ["6011", "6127"], ["7021", "7165"], ["12959", "13948"]],
        [
            {"type": "variant", "description": "-10del", "start": 25, "end": 26},
        ],
    ),
]


@pytest.mark.parametrize("payload, exons, expected", view_variants)
def test_parse_view_variants(
    payload: List[Dict[str, Any]], exons: List[List[str]], expected: List[Any]
) -> None:
    # exons = [["5027", "5113"], ["6011", "6127"], ["7021", "7165"], ["12959", "13948"]]
    assert parse_view_variants(exons, payload) == expected


variants = [
    # exon: Range, variants, expected
    ((0, 10), [{"start": 100}], list()),
    (
        (0, 10),
        [{"start": 0, "description": "274G>T"}],
        [Variant(0, "274G>T", color="red")],
    ),
    (
        (100, 110),
        [{"start": 105, "description": "274G>T"}],
        [Variant(5, "274G>T", color="red")],
    ),
]


@pytest.mark.parametrize("exon, variants, expected", variants)
def test_exon_variants(
    exon: Range, variants: List[Dict[str, Any]], expected: List[Variant]
) -> None:
    assert exon_variants(exon, variants) == expected


inside_exon_variants = [
    ({"start": 0}, True),
    ({"start": 10}, False),
    ({"start": -1}, False),
]


@pytest.mark.parametrize("variant, expected", inside_exon_variants)
def test_variant_inside_exon(variant: Dict[str, Any], expected: bool) -> None:
    exon = (0, 10)
    assert inside(exon, variant) == expected


# def test_convert_mutalyzer_range() -> None:
#     assert convert_mutalyzer_range("238", "11295") == (237, 11294)
#     assert convert_mutalyzer_range("29199", "7218") == (29198, 7217)


def test_exons_to_ranges() -> None:
    exons_in = [["100", "199"], ["300", "349"]]
    exons_out = [(0, 100), (100, 150)]
    assert exons_to_ranges(exons_in, list()) == exons_out


def test_exons_to_ranges_reverse() -> None:
    exons_in = [["349", "300"], ["199", "100"]]
    exons_out = [(0, 50), (50, 150)]
    assert exons_to_ranges(exons_in, list()) == exons_out


def test_cds_to_ranges() -> None:
    exons_in = [["100", "199"], ["300", "349"]]
    cds_in = ["150", "320"]
    cds_out = (50, 121)
    assert cds_to_ranges(exons_in, cds_in) == cds_out


def test_cds_to_ranges_reverse() -> None:
    exons_in = [["349", "300"], ["199", "100"]]
    cds_in = ["320", "150"]
    cds_out = (29, 100)
    assert cds_to_ranges(exons_in, cds_in) == cds_out


def test_variant_to_ranges() -> None:
    exons_in = [["100", "199"], ["300", "349"]]
    variant_start = 150
    variant_end = 319
    expected_start = 51
    expected_end = 120
    assert variant_to_ranges(exons_in, variant_start, variant_end) == (
        expected_start,
        expected_end,
    )


def test_variant_to_ranges_reverse() -> None:
    exons_in = [["349", "300"], ["199", "100"]]
    variant_start = 150
    variant_end = 319
    expected_start = 29
    expected_end = 98
    assert variant_to_ranges(exons_in, variant_start, variant_end) == (
        expected_start,
        expected_end,
    )


variants_exons = [
    (
        # Exons
        [["4952", "4794"], ["3616", "2954"]],
        # Variant
        {"type": "variant", "description": "130del", "start": 3615, "end": 3614},
        # is exonic
        True,
    ),
    (
        [["5027", "5113"], ["6011", "6127"], ["7021", "7165"], ["12959", "13948"]],
        {"type": "variant", "description": "274G>T", "start": 7124, "end": 7125},
        True,
    ),
    (
        [["5027", "5113"], ["6011", "6127"], ["7021", "7165"], ["12959", "13948"]],
        {"type": "variant", "description": "53del", "start": 6010, "end": 6011},
        True,
    ),
    # Intronic variant in NG_012337.3(NM_003002.4), before the exon
    (
        [["5027", "5113"], ["6011", "6127"], ["7021", "7165"], ["12959", "13948"]],
        {"type": "variant", "description": "53-10del", "start": 6000, "end": 6001},
        False,
    ),
    # Intronic variant in NG_012337.3(NM_003002.4), after the exon
    (
        [["5027", "5113"], ["6011", "6127"], ["7021", "7165"], ["12959", "13948"]],
        {"type": "variant", "description": "52+10del", "start": 5122, "end": 5123},
        False,
    ),
    # Variant before the transcript start of NG_012337.3(NM_003002.4)
    (
        [["5027", "5113"], ["6011", "6127"], ["7021", "7165"], ["12959", "13948"]],
        {"type": "variant", "description": "-50del", "start": 5011, "end": 5012},
        False,
    ),
    # Non-coding variant in NG_012337.3(NM_003002.4), inside an exon
    (
        [["5027", "5113"], ["6011", "6127"], ["7021", "7165"], ["12959", "13948"]],
        {"type": "variant", "description": "-10del", "start": 5051, "end": 5052},
        True,
    ),
]


@pytest.mark.parametrize("exons, variant, expected", variants_exons)
def test_detect_exon_variant(
    exons: List[List[str]], variant: Dict[str, Any], expected: bool
) -> None:
    assert exon_variant(exons, variant) == expected


rewrite_rev_vars = [
    # view_variant, after_rewrite
    # Forward, we don't do anything
    (
        {"views": [{"start": 0, "end": 10}]},
        {"views": [{"start": 0, "end": 10}]},
    ),
    # Reverse, we update the view postions
    (
        {
            "seq_length": 15948,
            "inverted": True,
            "views": [
                {"start": 0, "end": 12332},
                {"start": 12332, "end": 12333},
                {"start": 12333, "end": 15948},
            ],
        },
        {
            "seq_length": 15948,
            "inverted": True,
            "views": [
                {"start": 15947, "end": 3615},
                {"start": 3615, "end": 3614},
                {"start": 3614, "end": -1},
            ],
        },
    ),
]


@pytest.mark.parametrize("view_variants, expected", rewrite_rev_vars)
def test_rewrite_reverse_variants(
    view_variants: Dict[str, Any], expected: Dict[str, Any]
) -> None:
    """Test rewriting variant payload on the reverse strand"""
    rewrite_reverse_variants(view_variants)
    assert view_variants == expected
